#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of Rok.
#
# Copyright © 2020-2022 Arrikto Inc.  All Rights Reserved.

"""Script to create a ServiceAccount for an external client.

This script aids on getting or creating a ServiceAccount in a specific
namespace and returns its long-lived token. The token can be used in an
external client for authentication purposes.
"""

import os
import sys
import stat
import time
import base64
import logging

from rok_kubernetes import config, models
from rok_kubernetes.exceptions import ApiException

from rok_common.encoding import force_unicode
from rok_common.fileutils import dump_to_file

from rok_kubernetes.client import (RoleClient, SecretClient, RoleBindingClient,
                                   ServiceAccountClient)

from rok_tasks import frontend, question
from rok_tasks.frontend.cli import RokCLI

log = logging.getLogger(__name__)
fr = frontend.get_frontend()

TITLE = "Kubernetes ServiceAccount Creator"
DESCRIPTION = "Generate a ServiceAccount on Kubernetes"

# Questions
######################################################


class ServiceAccountName(question.LineInputQuestion):
    """Question for Service Account Name."""

    name = "question/service_account_name"
    msg = "What is the name of the ServiceAccount?"
    required = True
    argument_opts = {
        "name": ["--service-account-name"],
        "metavar": "SERVICE_ACCOUNT_NAME",
        "help": "Use ServiceAccount with name %(metavar)s"
    }


class ServiceAccountNamespace(question.LineInputQuestion):
    """Question for Service Account Namespace."""

    def __init__(self, *args, **kwargs):
        super(ServiceAccountNamespace, self).__init__(*args, **kwargs)
        sa_path = "/var/run/secrets/kubernetes.io/serviceaccount/namespace"
        if os.path.exists(sa_path):
            with open(sa_path) as f:
                self.default = f.readline()

    name = "question/service_account_namespace"
    msg = "What is the namespace of the ServiceAccount?"
    required = True
    argument_opts = {
        "name": ["--service-account-namespace"],
        "metavar": "SERVICE_ACCOUNT_NAMESPACE",
        "help": "Use ServiceAccount with namespace %(metavar)s"
    }


# Helpers
######################################################


def load_kube_config():
    # Load K8s configuration, if present
    if "KUBERNETES_SERVICE_HOST" in os.environ:
        config.load_incluster_config()
    else:
        try:
            config.load_kube_config()
        except IOError as e:
            raise RuntimeError("No K8s configuration was found: %s" % e)


def get_or_create_service_account(body):
    name, namespace = body.metadata.name, body.metadata.namespace
    sa_client = ServiceAccountClient()
    log.info("Checking if the ServiceAccount '%s' exists in namespace '%s'..."
             % (name, namespace))
    try:
        sa = sa_client.get(name, namespace=namespace)
        log.info("ServiceAccount already exists")
    except ApiException as e:
        if e.status == 404:
            sa = sa_client.create(body, namespace=namespace)
    return sa


def get_sa_token_role(sa):
    meta = models.V1ObjectMeta(name="%s-create-token" % sa.metadata.name,
                               namespace=sa.metadata.namespace)
    rule = models.V1PolicyRule(api_groups=[""],
                               resources=["serviceaccounts/token"],
                               resource_names=[sa.metadata.name],
                               verbs=["create"])
    return models.V1Role(api_version="rbac.authorization.k8s.io/v1",
                         kind="Role", metadata=meta, rules=[rule])


def get_role_binding(sa, role):
    meta = models.V1ObjectMeta(name="%s-create-token" % sa.metadata.name,
                               namespace=sa.metadata.namespace)
    role_ref = models.V1RoleRef(api_group="rbac.authorization.k8s.io",
                                kind="Role", name=role.metadata.name)
    subject = models.V1Subject(kind="ServiceAccount", name=sa.metadata.name,
                               namespace=sa.metadata.namespace)
    return models.V1RoleBinding(api_version="rbac.authorization.k8s.io/v1",
                                kind="RoleBinding", metadata=meta,
                                role_ref=role_ref, subjects=[subject])


def get_sa_token(sa):
    if not sa.secrets:
        raise ValueError("No tokens exist for the specific ServiceAccount")
    secret_name = sa.secrets[0].name
    secret = SecretClient().get(secret_name, namespace=sa.metadata.namespace)
    if secret.data.get("token") is None:
        msg = "The ServiceAccount secret is not populated with a token"
        raise RuntimeError(msg)
    return base64.b64decode(secret.data.get("token"))


def authorize_token_creation(sa):
    role_client = RoleClient()
    role, _ = role_client.get_or_create(get_sa_token_role(sa),
                                        namespace=sa.metadata.namespace)
    rb_client = RoleBindingClient()
    rb_client.get_or_create(get_role_binding(sa, role),
                            namespace=sa.metadata.namespace)


# CLI
######################################################


class ServiceAccountCreator(RokCLI):
    """Create a ServiceAccount and return its long-lived token."""

    IDENTIFIABLE_LABELS = {"arrikto.com/service-account-creator": ""}
    DEFAULT_LOGFILE = "service_account_creator.log"
    DEFAULT_FRONTEND = "readline"

    def initialize_parser(self):
        parser = super(ServiceAccountCreator, self).initialize_parser()
        questions = [ServiceAccountName(), ServiceAccountNamespace()]
        self.questions = questions
        question.add_questions_args(questions, parser)
        return parser

    def _ask_questions(self):
        for q in self.questions:
            self.qctx.add_question(q)
            q.ask()
        return

    def make_service_account(self):
        sa_name = self.qctx.questions[ServiceAccountName.name].get_answer()
        sa_namespace = (self.qctx.questions[ServiceAccountNamespace.name]
                        .get_answer())
        meta = models.V1ObjectMeta(name=sa_name, namespace=sa_namespace,
                                   labels=self.IDENTIFIABLE_LABELS)
        return models.V1ServiceAccount(api_version="v1", kind="ServiceAccount",
                                       metadata=meta)

    def main(self):
        load_kube_config()
        self._ask_questions()

        sa_name = self.qctx.questions[ServiceAccountName.name].get_answer()
        sa_namespace = (self.qctx.questions[ServiceAccountNamespace.name]
                        .get_answer())

        with fr.Progress(maxval=4, title=TITLE) as p:

            p.info("Creating ServiceAccount...")
            sa = self.make_service_account()
            sa = get_or_create_service_account(sa)
            p.inc()
            p.success("Created ServiceAccount!")

            p.info("Authorizing ServiceAccount to create tokens for itself...")
            authorize_token_creation(sa)
            p.inc()
            p.success("Authorized ServiceAccount!")

            p.info("Getting a long-lived token for the ServiceAccount...")
            # We need to wait for the token controller to issue a token and
            # update the ServiceAccount. Then, we fetch the service account
            # again and retrieve its token
            time.sleep(1)
            token = get_sa_token(get_or_create_service_account(sa))
            p.inc()
            p.success("Retrieved token!")

            p.info("Storing token in file...")
            file_name = "%s-%s.token" % (sa_namespace, sa_name)
            mode = stat.S_IRUSR | stat.S_IWUSR
            with dump_to_file(file_name, mode=mode) as f:
                f.write(force_unicode(token))
            p.inc()

            p.success("Successfully created ServiceAccount `%s/%s' and"
                      " retrieved its token!"
                      % (sa_namespace, sa_name))

        msg = (u"Successfully retrieved ServiceAccount `%s/%s' token.\n"
               u"Token is stored in file `%s'." % (sa_namespace, sa_name,
                                                   file_name))

        with fr.Message(msg, title=TITLE, lvl=frontend.DEBUG,
                        ok_button=True) as m:
            m.show()


def main():
    cli = ServiceAccountCreator()
    return cli.run()


if __name__ == "__main__":
    sys.exit(main())
